""" Elements are mini chameleon macros, usually used to render a single
element. Sometimes it is useful to not just apply a macro, but to have an
actual object representing that code snippet. The prime example being links.

Elements are rendered in chameleon templates by calling them with the layout
object::

    <tal:b replace="structure element(layout)" />

Note that no matter how quick the elements rendering is, using chameleon
templates directly is faster.

This module should eventually replace the elements.py module.

"""

from onegov.core.templates import render_macro
from onegov.org import _

# currently used for backwards compatibility, this should become a regular
# element as well
from onegov.org.elements import LinkGroup  # noqa


class Element(object):
    """ Provides a way to define elements trough multiple variables. These
    elements may then be rendered by the elements.pt templates.

    Elements are simple, they render whatever text and attributes the get.
    They also carry a dictionary with properties on them, so any kind of
    value may be attached to an element. For example this works::

        user = object()
        element = Element(model=user)
        assert element.user == user

    Since we might use a complex set of attributes which need to be set in
    a certain way we use a traits system to avoid having to rely too heavily
    on class inheritance.

    Basically traits are callables called with attributes which they may
    manipulate. For example, a trait might turn a simple argument into
    a set of attributes::

        Element(traits=[Editor(buttons=['new', 'edit', 'delete'])])

    See the link traits further down to see examples.

    """

    # The link refers to the id of the macro written in elements.pt
    id = None

    __slots__ = ('text', 'attrs', 'props')

    def __init__(self, text=None, attrs=None, traits=(), **props):
        self.text = text
        self.props = props
        self.attrs = self.normalize_attrs(attrs)

        if isinstance(traits, Trait):
            traits = (traits, )

        for trait in traits:
            self.attrs = trait(self.attrs, **props)

    def normalize_attrs(self, attrs):
        """ Before we do anything with attributes we make sure we can easily
        add/remove classes from them, so we use a set.

        """
        attrs = attrs or {}

        if 'class' not in attrs:
            attrs['class'] = set()
        elif isinstance(attrs['class'], str):
            attrs['class'] = set(attrs['class'].split(' '))
        else:
            attrs['class'] = set(attrs['class'])

        return attrs

    def __eq__(self, other):
        return self.id == other.id\
            and self.attrs == other.attrs\
            and self.props == other.props

    def __getattr__(self, name):
        if name in self.props:
            return self.props[name]

        raise AttributeError(name)

    def __call__(self, layout):
        if self.attrs['class']:
            self.attrs['class'] = ' '.join(self.attrs['class'])
        else:
            del self.attrs['class']

        return render_macro(layout.elements[self.id], layout.request, {
            'e': self, 'layout': layout,
        })


class HiddenLinkMixin(object):
    """ Hidden links point to views which are not available to the public
    (usually through a publication mechanism).

    We mark those links with an icon. This mixin automatically does that
    for links which bear a model property.

    """

    __slots__ = ()

    @property
    def is_hidden_from_public(self):
        """ Returns True if Link is hidden from the public. Pass extra keyword
        ``model`` to ``__init__`` to have this work automatically.

        """
        if hasattr(self, 'model'):
            return getattr(self.model, 'is_hidden_from_public', False)

        return False


class Link(Element, HiddenLinkMixin):
    """ A generic link. """

    id = 'link'

    __slots__ = ()

    def __init__(self, text, url='#', attrs=None, traits=(), **props):
        # this is the only exception we permit where we don't use a trait
        # to change the attributes - we do this because the url is essential
        # to a link, so it makes sense to have it as a proper argument
        attrs = attrs or {}
        attrs['href'] = url

        super().__init__(text, attrs, traits, **props)


class Trait(object):
    """ Base class for all traits. """

    __slots__ = ('apply', )

    def __call__(self, attrs):
        return self.apply(attrs)


class Confirm(Trait):
    """ Links with a confirm link will only be triggered after a question has
    been answered with a yes::

        link = Link(_("Continue"), '/last-step', traits=(
            Confirm(
                _("Do you really want to continue?"),
                _("This cannot be undone"),
                _("Yes"), _("No")
            ),
        ))

    """

    __slots__ = ()

    def __init__(self, confirm, extra=None, yes=_("Yes"), no=_("Cancel"),
                 **ignored):
        def apply(attrs):
            attrs['class'].add('confirm')
            attrs['data-confirm'] = confirm
            attrs['data-confirm-extra'] = extra
            attrs['data-confirm-yes'] = yes
            attrs['data-confirm-no'] = no
            return attrs

        self.apply = apply


class Block(Confirm):
    """ A block trait is a confirm variation with no way to say yes. """

    __slots__ = ()

    def __init__(self, message, extra=None, **ignored):
        super().__init__(message, extra, yes=None)


class Intercooler(Trait):
    """ Turns the link into an intercooler request. As a result, the link
    will be exeucted as an ajax request.

    See http://intercoolerjs.org.
    """

    __slots__ = ()

    def __init__(self, request_method, target=None, redirect_after=None,
                 **ignored):
        def apply(attrs):
            if request_method == 'GET':
                attrs['ic-get-from'] = attrs['href']
            elif request_method == 'POST':
                attrs['ic-post-to'] = attrs['href']
            elif request_method == 'DELETE':
                attrs['ic-delete-from'] = attrs['href']

            del attrs['href']
            attrs['ic-target'] = target

            if redirect_after:
                attrs['redirect-after'] = redirect_after

            return attrs

        self.apply = apply



from dectate import Action


class HomepageWidgetAction(Action):
    """ Register a cronjob. """

    config = {
        'homepage_widget_registry': dict
    }

    def __init__(self, tag):
        self.tag = tag

    def identifier(self, homepage_widget_registry):
        return self.tag

    def perform(self, func, homepage_widget_registry):
        widget = func()
        widget.tag = self.tag  # keep redundantly for ease of access

        homepage_widget_registry[self.tag] = widget


class ExportAction(Action):
    """ Register an export. """

    config = {
        'export_registry': dict
    }

    def __init__(self, id, **kwargs):
        self.id = id
        self.kwargs = kwargs
        self.kwargs['id'] = id

    def identifier(self, export_registry):
        return self.id

    def perform(self, cls, export_registry):
        export_registry[self.id] = cls(**self.kwargs)



from onegov.core.utils import Bunch
from onegov.org import _
from onegov.org.new_elements import Link, Confirm, Intercooler


def test_link(render_element):
    # text is translated
    result = render_element(Link(text=_("Settings"), url='/settings'))
    assert result.pyquery('a').text() == "Einstellungen"
    assert result.pyquery('a').attr('href') == '/settings'

    # other attributes are rendered
    result = render_element(Link(text='foo', url='#', attrs={
        'data-foo': 'bar'
    }))
    assert result.pyquery('a').attr('data-foo') == 'bar'

    # we show a hint if the link is hidden from public
    result = render_element(Link(text='hidden', url='#', model=Bunch(
        is_hidden_from_public=True
    )))

    assert result.pyquery('i').attr('title') \
        == "Diese Seite ist nicht öffentlich."
    assert result.pyquery('i').attr('class') == 'hidden-from-public-hint'


def test_confirm_link(render_element):
    result = render_element(Link(text="Delete", url='#', traits=(
        Confirm(
            _("Confirm?"),
            _("Extra..."),
            _("Yes"),
            _("No")
        ),
    ), attrs={'class': 'foo'}))

    assert result.pyquery('a').attr('data-confirm') == "Confirm?"
    assert result.pyquery('a').attr('data-confirm-extra') == "Extra..."
    assert result.pyquery('a').attr('data-confirm-yes') == "Ja"
    assert result.pyquery('a').attr('data-confirm-no') == "Nein"
    assert result.pyquery('a').attr('class') in ('foo confirm', 'confirm foo')


def test_link_slots():
    # make sure that the Link class as well as all its parents have
    # __slots__ defined (for some lookup speed and memory improvements)
    assert not hasattr(Link("Slots", '#'), '__dict__')


def test_intercooler_link(render_element):
    result = render_element(Link(text="Delete", traits=Intercooler(
        request_method="POST", redirect_after='#redirect', target='#target'
    )))

    assert result.pyquery('a').attr('ic-post-to') == '#'
    assert result.pyquery('a').attr('ic-target') == '#target'
    assert result.pyquery('a').attr('redirect-after') == '#redirect'
    assert result.pyquery('a').attr('href') is None


def test_class_attributes(render_element):
    result = render_element(Link(text="Delete", attrs={
        'class': 'foo'
    }))
    assert result.pyquery('a').attr('class') == 'foo'

    result = render_element(Link(text="Delete", attrs={
        'class': ('foo', 'bar')
    }))
    assert result.pyquery('a').attr('class') in ('foo bar', 'bar foo')

    result = render_element(Link(text="Delete", attrs={
        'class': ('foo', 'bar')
    }))
    assert result.pyquery('a').attr('class') in ('foo bar', 'bar foo')

    result = render_element(Link(text="Delete"))
    assert result.pyquery('a').attr('class') is None



import morepath
import sedate

from collections import OrderedDict, namedtuple
from datetime import datetime, timedelta
from isodate import parse_date, ISO8601Error
from itertools import groupby
from libres.db.models import Reservation
from onegov.core.security import Public, Private
from onegov.form import FormSubmission
from onegov.reservation import ResourceCollection
from onegov.reservation.models import Resource
from onegov.org import _, OrgApp, utils
from onegov.org.elements import Link
from onegov.org.forms import (
    ResourceForm, ResourceCleanupForm, ResourceExportForm
)
from onegov.org.layout import ResourcesLayout, ResourceLayout
from onegov.org.models.resource import DaypassResource, RoomResource
from onegov.ticket import Ticket
from sqlalchemy.sql.expression import nullsfirst
from webob import exc


RESOURCE_TYPES = {
    'daypass': {
        'success': _("Added a new daypass"),
        'title': _("New daypass"),
        'class': DaypassResource
    },
    'room': {
        'success': _("Added a new room"),
        'title': _("New room"),
        'class': RoomResource
    }
}


def get_daypass_form(self, request):
    return get_resource_form(self, request, 'daypass')


def get_room_form(self, request):
    return get_resource_form(self, request, 'room')


def get_resource_form(self, request, type=None):
    if isinstance(self, ResourceCollection):
        assert type is not None
        model = RESOURCE_TYPES[type]['class']()
    else:
        model = self

    return model.with_content_extensions(ResourceForm, request)


def get_grouped_resources(resources, request, transform=None, emptygroup=None):
    resources = resources.query().order_by(nullsfirst(Resource.group))
    resources = request.exclude_invisible(resources)

    grouped = OrderedDict()

    def group_key(resource):
        return resource.group or request.translate(_("General"))

    def sort_key(resource):
        return resource.title

    transform = transform or (lambda value: value)

    for group, items in groupby(resources, group_key):
        grouped[group] = sorted([i for i in items], key=sort_key)
        grouped[group] = [transform(i) for i in grouped[group]]

    if len(grouped) == 1:
        grouped = {emptygroup: tuple(grouped.values())[0]}

    return grouped


@OrgApp.html(model=ResourceCollection, template='resources.pt',
             permission=Public)
def view_resources(self, request):
    return {
        'title': _("Reservations"),
        'resources': get_grouped_resources(self, request),
        'layout': ResourcesLayout(self, request)
    }


@OrgApp.json(model=ResourceCollection, permission=Public, name='json')
def view_resources_json(self, request):

    def transform(resource):
        return {
            'name': resource.name,
            'title': resource.title,
            'url': request.link(resource),
        }

    @request.after
    def cache(response):
        # only update once every minute
        response.cache_control.max_age = 60

    return get_grouped_resources(
        self, request,
        transform=transform,
        emptygroup=request.translate(_("Reservations"))
    )


@OrgApp.form(model=ResourceCollection, name='neuer-raum',
             template='form.pt', permission=Private, form=get_room_form)
def handle_new_room(self, request, form):
    return handle_new_resource(self, request, form, 'room')


@OrgApp.form(model=ResourceCollection, name='neue-tageskarte',
             template='form.pt', permission=Private, form=get_daypass_form)
def handle_new_daypass(self, request, form):
    return handle_new_resource(self, request, form, 'daypass')


def handle_new_resource(self, request, form, type):
    if form.submitted(request):

        resource = self.add(
            title=form.title.data, type=type, timezone='Europe/Zurich'
        )
        form.populate_obj(resource)

        request.success(RESOURCE_TYPES[type]['success'])
        return morepath.redirect(request.link(resource))

    layout = ResourcesLayout(self, request)
    layout.include_editor()
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(RESOURCE_TYPES[type]['title'], '#'))

    return {
        'layout': layout,
        'title': _(RESOURCE_TYPES[type]['title']),
        'form': form,
        'form_width': 'large'
    }


@OrgApp.form(model=Resource, name='bearbeiten', template='form.pt',
             permission=Private, form=get_resource_form)
def handle_edit_resource(self, request, form):
    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))
        return morepath.redirect(request.link(self))

    elif not request.POST:
        form.process(obj=self)

    layout = ResourceLayout(self, request)
    layout.include_editor()
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(_("Edit"), '#'))

    return {
        'layout': layout,
        'title': self.title,
        'form': form,
        'form_width': 'large'
    }


@OrgApp.html(model=Resource, template='resource.pt', permission=Public)
def view_resource(self, request):
    return {
        'title': self.title,
        'resource': self,
        'layout': ResourceLayout(self, request),
        'feed': request.link(self, name='slots'),
        'resources_url': request.class_link(ResourceCollection, name='json')
    }


@OrgApp.view(model=Resource, request_method='DELETE', permission=Private)
def handle_delete_resource(self, request):

    request.assert_valid_csrf_token()

    if not self.deletable:
        raise exc.HTTPMethodNotAllowed()

    collection = ResourceCollection(request.app.libres_context)
    collection.delete(self, including_reservations=False)


@OrgApp.form(model=Resource, permission=Private, name='cleanup',
             form=ResourceCleanupForm, template='resource_cleanup.pt')
def handle_cleanup_allocations(self, request, form):
    """ Removes all unused allocations between the given dates. """

    if form.submitted(request):
        start, end = form.data['start'], form.data['end']
        count = self.scheduler.remove_unused_allocations(start, end)

        request.success(
            _("Successfully removed ${count} unused allocations", mapping={
                'count': count
            })
        )

        return morepath.redirect(request.link(self))

    if request.method == 'GET':
        form.start.data, form.end.data = get_date_range(self, request.params)

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("Clean up"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("Clean up"),
        'form': form
    }


def predict_next_reservation(resource, request, reservations):

    prediction = utils.predict_next_daterange(
        tuple((r.display_start(), r.display_end()) for r in reservations)
    )

    if not prediction:
        return None

    allocation = resource.scheduler.allocations_in_range(*prediction).first()

    if not allocation:
        return None

    whole_day = sedate.is_whole_day(*prediction, timezone=resource.timezone)
    quota = utils.predict_next_value(tuple(r.quota for r in reservations)) or 1

    if whole_day:
        time = request.translate(_("Whole day"))
    else:
        time = utils.render_time_range(*prediction)

    return {
        'url': request.link(allocation, name='reserve'),
        'start': prediction[0].isoformat(),
        'end': prediction[1].isoformat(),
        'quota': quota,
        'wholeDay': whole_day,
        'time': time
    }


@OrgApp.json(model=Resource, name='reservations', permission=Public)
def get_reservations(self, request):

    reservations = tuple(self.bound_reservations(request))
    prediction = predict_next_reservation(self, request, reservations)

    return {
        'reservations': [
            utils.ReservationInfo(reservation, request).as_dict()
            for reservation in reservations
        ],
        'prediction': prediction
    }


def get_date(text, default):
    try:
        date = parse_date(text)
        return datetime(date.year, date.month, date.day, tzinfo=default.tzinfo)
    except (ISO8601Error, TypeError):
        return default


def get_date_range(resource, params):
    default_start, default_end = resource.calendar_date_range

    start = get_date(params.get('start'), default_start)
    end = get_date(params.get('end'), default_end)

    start = sedate.replace_timezone(
        datetime(start.year, start.month, start.day), resource.timezone)

    end = sedate.replace_timezone(
        datetime(end.year, end.month, end.day), resource.timezone)

    if end < start:
        start = end

    return sedate.align_range_to_day(start, end, resource.timezone)


@OrgApp.html(model=Resource, permission=Private, name='belegung',
             template='resource_occupancy.pt')
def view_occupancy(self, request):

    # infer the default start/end date from the calendar view parameters
    start, end = get_date_range(self, request.params)

    query = self.reservations_with_tickets_query(start, end)
    query = query.with_entities(
        Reservation.start, Reservation.end, Reservation.quota,
        Ticket.subtitle, Ticket.id
    )

    def group_key(record):
        return sedate.to_timezone(record[0], self.timezone).date()

    occupancy = OrderedDict()
    grouped = groupby(query.all(), group_key)
    Entry = namedtuple('Entry', ('start', 'end', 'title', 'quota', 'url'))
    count = 0

    for date, records in grouped:
        occupancy[date] = tuple(
            Entry(
                start=sedate.to_timezone(r[0], self.timezone),
                end=sedate.to_timezone(
                    r[1] + timedelta(microseconds=1), self.timezone),
                quota=r[2],
                title=r[3],
                url=request.class_link(Ticket, {
                    'handler_code': 'RSV',
                    'id': r[4]
                })
            ) for r in records
        )
        count += len(occupancy[date])

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("Occupancy"), '#'))
    layout.editbar_links = None

    utilisation = 100 - self.scheduler.queries.availability_by_range(
        start, end, (self.id, )
    )

    return {
        'layout': layout,
        'title': _("Occupancy"),
        'occupancy': occupancy,
        'resource': self,
        'start': sedate.to_timezone(start, self.timezone).date(),
        'end': sedate.to_timezone(end, self.timezone).date(),
        'count': count,
        'utilisation': utilisation
    }


@OrgApp.form(model=Resource, permission=Private, name='export',
             template='export.pt', form=ResourceExportForm)
def view_export(self, request, form):

    layout = ResourceLayout(self, request)
    layout.breadcrumbs.append(Link(_("Occupancy"), '#'))
    layout.editbar_links = None

    # XXX this could be turned into a redirect to a GET view, which would
    # make it easier for scripts to get this data, but since we don't have
    # a good API story anyway we don't have spend to much energy on it here
    # - instead we should do this in a comprehensive fashion
    if form.submitted(request):
        constant_fields, results = run_export(
            resource=self,
            request=request,
            start=form.data['start'],
            end=form.data['end'],
            nested=form.format == 'json',
            formatter=layout.export_formatter(form.format)
        )

        def field_order(field):
            # known fields come first in a defined order (-x -> -1)
            # unknown fields are ordered from a-z (second item in tuple)
            if field in constant_fields:
                return constant_fields.index(field) - len(constant_fields), ''
            else:
                return 0, field

        return form.as_export_response(results, self.title, key=field_order)

    if request.method == 'GET':
        form.start.data, form.end.data = get_date_range(self, request.params)

    return {
        'layout': layout,
        'title': _("Export"),
        'form': form,
        'explanation': _("Exports the reservations of the given date range.")
    }


def run_export(resource, request, start, end, nested, formatter):
    start = sedate.replace_timezone(
        datetime(start.year, start.month, start.day),
        resource.timezone
    )
    end = sedate.replace_timezone(
        datetime(end.year, end.month, end.day),
        resource.timezone
    )

    start, end = sedate.align_range_to_day(start, end, resource.timezone)

    query = resource.reservations_with_tickets_query(start, end)
    query = query.join(FormSubmission, Reservation.token == FormSubmission.id)
    query = query.with_entities(
        Reservation.start,
        Reservation.end,
        Reservation.quota,
        Reservation.email,
        Ticket.number,
        Ticket.subtitle,
        FormSubmission.data,
    )

    results = []

    # update me: reused outside this function
    constant_fields = ('start', 'end', 'quota', 'email', 'ticket', 'title')

    for record in query:
        result = OrderedDict()

        start = sedate.to_timezone(record[0], resource.timezone)
        end = sedate.to_timezone(record[1], resource.timezone)
        end += timedelta(microseconds=1)

        result['start'] = formatter(start)
        result['end'] = formatter(end)
        result['quota'] = formatter(record[2])
        result['email'] = formatter(record[3])
        result['ticket'] = formatter(record[4])
        result['title'] = formatter(record[5])

        if nested:
            result['form'] = record[6]
        else:
            for key, value in record[6].items():
                if isinstance(value, list):
                    result['form_' + key] = ', '.join(value)
                else:
                    result['form_' + key] = value

        results.append(result)

    return constant_fields, results



from onegov.form import Form, merge_forms
from onegov.form.validators import ValidFormDefinition
from onegov.org import _
from onegov.org.forms.fields import HtmlField
from onegov.org.forms.generic import DateRangeForm, ExportForm
from onegov.org.forms.reservation import RESERVED_FIELDS
from wtforms import StringField, TextAreaField, validators


class ResourceForm(Form):
    """ Defines the form for all resources. """
    title = StringField(_("Title"), [validators.InputRequired()])

    lead = TextAreaField(
        label=_("Lead"),
        description=_("Describes what this reservation resource is about"),
        render_kw={'rows': 4})

    group = StringField(
        label=_("Group"),
        description=_("Used to group the resource in the overview")
    )

    text = HtmlField(
        label=_("Text"))

    definition = TextAreaField(
        label=_("Extra Fields Definition"),
        validators=[
            validators.Optional(),
            ValidFormDefinition(
                require_email_field=False,
                reserved_fields=RESERVED_FIELDS
            )
        ],
        render_kw={'rows': 32, 'data-editor': 'form'}
    )


class ResourceCleanupForm(DateRangeForm):
    """ Defines the form to remove multiple allocations. """


class ResourceExportForm(merge_forms(DateRangeForm, ExportForm)):
    """ Resource export form with start/end date. """




import os

from onegov.core.utils import module_path, rchop
from onegov.event import EventCollection, OccurrenceCollection
from onegov.form import FormCollection
from onegov.reservation import ResourceCollection
from onegov.page import PageCollection


def test_initial_content(org_app):
    pages = PageCollection(org_app.session()).query().all()
    pages = {p.name: p.title for p in pages}

    assert pages == {
        'organisation': 'Organisation',
        'themen': 'Themen',
        'kontakt': 'Kontakt',
        'aktuelles': 'Aktuelles',
        'wir-haben-eine-neue-webseite': 'Wir haben eine neue Webseite!'
    }

    forms = FormCollection(org_app.session()).definitions.query().all()
    forms = set(form.name for form in forms)

    builtin_forms_path = module_path('onegov.org', 'forms/builtin')

    paths = (p for p in os.listdir(builtin_forms_path))
    paths = (p for p in paths if p.endswith('.form'))
    paths = (os.path.basename(p) for p in paths)
    builtin_forms = set(rchop(p, '.form') for p in paths)

    assert builtin_forms == forms

    resources = ResourceCollection(org_app.libres_context).query().all()
    resources = {r.name: r.type for r in resources}

    assert resources == {
        'konferenzraum': 'room',
        'tageskarte': 'daypass',
    }

    assert EventCollection(org_app.session()).query().count() == 4
    assert OccurrenceCollection(org_app.session()).query().count() > 4



import json
import morepath
import sedate
import transaction

from datetime import time
from libres.db.models import Allocation, Reservation
from libres.modules.errors import LibresError
from onegov.core.security import Public, Private
from onegov.form import FormCollection, merge_forms
from onegov.reservation import Resource
from onegov.org import utils
from onegov.org.elements import Link
from onegov.org.forms import ReservationForm
from onegov.org.layout import ReservationLayout
from onegov.ticket import TicketCollection
from onegov.org import _, OrgApp
from onegov.org.mail import send_html_mail
from purl import URL
from sqlalchemy.orm.attributes import flag_modified
from webob import exc


def assert_anonymous_access_only_temporary(resource, reservation, request):
    """ Raises exceptions if the current user is anonymous and no longer
    should be given access to the reservation models.

    This could probably be done using morepath's security system, but it would
    not be quite as straight-forward. This approach is, though we have
    to manually add this function to all reservation views the anonymous user
    should be able to access when creating a new reservatin, but not anymore
    after that.

    """
    if request.is_manager:
        return

    if not reservation.session_id:
        raise exc.HTTPForbidden()

    if reservation.status == 'approved':
        raise exc.HTTPForbidden()

    if reservation.session_id != resource.bound_session_id(request):
        raise exc.HTTPForbidden()


def assert_access_only_if_there_are_reservations(reservations):
    """ Raises an exception if no reservations are available. """
    if not reservations:
        raise exc.HTTPForbidden()


@OrgApp.json(model=Allocation, name='reserve', request_method='POST',
             permission=Public)
def reserve_allocation(self, request):
    """ Adds a single reservation to the list of reservations bound to the
    current browser session.

    Does not actually reserve anything, just keeps a list of things to
    reserve later. Though it will still check if the reservation is
    feasable.

    """

    # the reservation is defined through query parameters
    start = request.params.get('start') or '{:%H:%M}'.format(self.start)
    end = request.params.get('end') or '{:%H:%M}'.format(self.end)
    quota = int(request.params.get('quota', 1))
    whole_day = request.params.get('whole_day') == '1'

    if self.partly_available:
        if self.whole_day and whole_day:
            start = time(0, 0)
            end = time(23, 59)
        else:
            start = sedate.parse_time(start)
            end = sedate.parse_time(end)

        start, end = sedate.get_date_range(
            sedate.to_timezone(self.start, self.timezone), start, end
        )
    else:
        start = self.start, self.end

    resource = request.app.libres_resources.by_allocation(self)

    try:
        resource.scheduler.reserve(
            email='0xdeadbeef@example.org',  # will be set later
            dates=(start, end),
            quota=quota,
            session_id=resource.bound_session_id(request),
            single_token_per_session=True
        )
    except LibresError as e:
        message = {
            'message': utils.get_libres_error(e, request),
            'success': False
        }

        @request.after
        def trigger_calendar_update(response):
            response.headers.add('X-IC-Trigger', 'rc-reservation-error')
            response.headers.add('X-IC-Trigger-Data', json.dumps(message))

        return message
    else:

        @request.after
        def trigger_calendar_update(response):
            response.headers.add('X-IC-Trigger', 'rc-reservations-changed')

        return {
            'success': True
        }


@OrgApp.json(model=Reservation, request_method='DELETE', permission=Public)
def delete_reservation(self, request):

    # anonymous users do not get a csrf token (it's bound to the identity)
    # therefore we can't check for it -> this is not a problem since
    # anonymous users do not really have much to lose here
    if request.is_logged_in:
        request.assert_valid_csrf_token()

    resource = request.app.libres_resources.by_reservation(self)

    # this view is public, but only for a limited time
    assert_anonymous_access_only_temporary(resource, self, request)

    try:
        resource.scheduler.remove_reservation(self.token, self.id)
    except LibresError as e:
        message = {
            'message': utils.get_libres_error(e, request),
            'success': False
        }

        @request.after
        def trigger_calendar_update(response):
            response.headers.add('X-IC-Trigger', 'rc-reservation-error')
            response.headers.add('X-IC-Trigger-Data', json.dumps(message))

        return message
    else:

        @request.after
        def trigger_calendar_update(response):
            response.headers.add('X-IC-Trigger', 'rc-reservations-changed')

        return {
            'success': True
        }


def get_reservation_form_class(resource, request):
    if resource.definition:
        return merge_forms(ReservationForm, resource.form_class)
    else:
        return ReservationForm


@OrgApp.form(model=Resource, name='formular', template='reservation_form.pt',
             permission=Public, form=get_reservation_form_class)
def handle_reservation_form(self, request, form):
    """ Asks the user for the form data required to complete one or many
    reservations on a resource.

    """
    reservations_query = self.bound_reservations(request)
    reservations = reservations_query.all()
    assert_access_only_if_there_are_reservations(reservations)

    # all reservations share a single token
    token = reservations[0].token

    # the submission shares the reservation token
    forms = FormCollection(request.app.session())
    submission = forms.submissions.by_id(token)

    # update the data if the form is submitted (even if invalid)
    if request.POST:
        # update the e-mail data
        reservations_query.order_by(False).update(
            {Reservation.email: form.email.data}
        )

        # while we re at it, remove all expired sessions
        self.remove_expired_reservation_sessions()

        # add the submission if it doesn't yet exist
        if self.definition and not submission:
            submission = forms.submissions.add_external(
                form=self.form_class(),
                state='pending',
                id=token
            )

        # update the data on the submission
        if submission:
            forms.submissions.update(
                submission, form, exclude=form.reserved_fields
            )

    # go to the next step if the submitted data is valid
    if form.submitted(request):
        return morepath.redirect(request.link(self, 'bestaetigung'))
    else:
        data = {}

        # the email is the same for all reservations
        if reservations[0].email != '0xdeadbeef@example.org':
            data['email'] = reservations[0].email

        if submission:
            data.update(submission.data)

        form.process(data=data)

    layout = ReservationLayout(self, request)
    layout.breadcrumbs.append(Link(_("Reserve"), '#'))

    title = _("New dates for ${title}", mapping={
        'title': self.title,
    })

    return {
        'layout': layout,
        'title': title,
        'form': form,
        'reservation_infos': [
            utils.ReservationInfo(r, request) for r in reservations
        ],
        'resource': self,
        'button_text': _("Continue")
    }


@OrgApp.html(model=Resource, name='bestaetigung', permission=Public,
             template='reservation_confirmation.pt')
def confirm_reservation(self, request):
    reservations = self.bound_reservations(request).all()
    assert_access_only_if_there_are_reservations(reservations)

    token = reservations[0].token.hex

    forms = FormCollection(request.app.session())
    submission = forms.submissions.by_id(token)

    if submission:
        form = request.get_form(submission.form_class, data=submission.data)
    else:
        form = None

    layout = ReservationLayout(self, request)
    layout.breadcrumbs.append(Link(_("Confirm"), '#'))

    failed_reservations = set(
        int(failed) for failed
        in request.params.get('failed_reservations', '').split(',')
        if failed
    )

    return {
        'title': _("Confirm your reservation"),
        'layout': layout,
        'form': form,
        'resource': self,
        'reservation_infos': [
            utils.ReservationInfo(r, request) for r in reservations
        ],
        'failed_reservations': failed_reservations,
        'finalize_link': request.link(self, 'abschluss'),
        'edit_link': request.link(self, 'formular'),
    }


@OrgApp.html(model=Resource, name='abschluss', permission=Public,
             template='layout.pt')
def finalize_reservation(self, request):
    reservations = self.bound_reservations(request).all()
    assert_access_only_if_there_are_reservations(reservations)

    session_id = self.bound_session_id(request)
    token = reservations[0].token.hex

    try:
        self.scheduler.queries.confirm_reservations_for_session(session_id)
        self.scheduler.approve_reservations(token)

    except LibresError as e:
        transaction.abort()
        utils.show_libres_error(e, request)

        url = URL(request.link(self, name='bestaetigung'))
        url = url.query_param('failed_reservations', e.reservation.id)

        return morepath.redirect(url.as_string())
    else:
        forms = FormCollection(request.app.session())
        submission = forms.submissions.by_id(token)

        if submission:
            forms.submissions.complete_submission(submission)

        with forms.session.no_autoflush:
            ticket = TicketCollection(request.app.session()).open_ticket(
                handler_code='RSV', handler_id=token
            )

        if reservations[0].email != request.current_username:
            send_html_mail(
                request=request,
                template='mail_ticket_opened.pt',
                subject=_("A ticket has been opened"),
                receivers=(reservations[0].email, ),
                content={
                    'model': ticket
                }
            )

        request.success(_("Thank you for your reservation!"))

        return morepath.redirect(request.link(ticket, 'status'))


@OrgApp.view(model=Reservation, name='annehmen', permission=Private)
def accept_reservation(self, request):
    if not self.data or not self.data.get('accepted'):
        resource = request.app.libres_resources.by_reservation(self)
        reservations = resource.scheduler.reservations_by_token(self.token)
        reservations = reservations.order_by(Reservation.start)

        if self.email != request.current_username:
            send_html_mail(
                request=request,
                template='mail_reservation_accepted.pt',
                subject=_("Your reservations were accepted"),
                receivers=(self.email, ),
                content={
                    'model': self,
                    'resource': resource,
                    'reservations': reservations
                }
            )

        for reservation in reservations:
            reservation.data = reservation.data or {}
            reservation.data['accepted'] = True

            # libres does not automatically detect changes yet
            flag_modified(reservation, 'data')

        request.success(_("The reservations were accepted"))
    else:
        request.warning(_("The reservations have already been accepted"))

    return request.redirect(request.link(self))


@OrgApp.view(model=Reservation, name='absagen', permission=Private)
def reject_reservation(self, request):
    resource = request.app.libres_resources.by_reservation(self)
    token = self.token.hex
    reservation_id = int(request.params.get('reservation-id', '0')) or None

    all_reservations = resource.scheduler.reservations_by_token(token)
    all_reservations = all_reservations.order_by(Reservation.start).all()

    targeted = tuple(r for r in all_reservations if r.id == reservation_id)
    targeted = targeted or all_reservations
    excluded = tuple(r for r in all_reservations if r.id not in {
        r.id for r in targeted
    })

    forms = FormCollection(request.app.session())
    submission = forms.submissions.by_id(token)

    if self.email != request.current_username:
        send_html_mail(
            request=request,
            template='mail_reservation_rejected.pt',
            subject=_("The following reservations were rejected"),
            receivers=(self.email, ),
            content={
                'model': self,
                'resource': resource,
                'reservations': targeted
            }
        )

    # create a snapshot of the ticket to keep the useful information
    if len(excluded) == 0:
        tickets = TicketCollection(request.app.session())
        ticket = tickets.by_handler_id(token)
        ticket.create_snapshot(request)

    for reservation in targeted:
        resource.scheduler.remove_reservation(token, reservation.id)

    if len(excluded) == 0 and submission:
        forms.submissions.delete(submission)

    if len(targeted) > 1:
        request.success(_("The reservations were rejected"))
    else:
        request.success(_("The reservation was rejected"))

    # return none on intercooler js requests
    if not request.headers.get('X-IC-Request'):
        return request.redirect(request.link(self))



from morepath.request import Response
from onegov.form import Form
from onegov.org import _
from onegov.core.csv import convert_list_of_dicts_to_csv
from onegov.core.csv import convert_list_of_dicts_to_xlsx
from onegov.core.utils import normalize_for_url
from wtforms import RadioField, validators
from wtforms.fields.html5 import DateField


class DateRangeForm(Form):
    """ A form providing a start/end date range. """

    start = DateField(
        label=_("Start"),
        validators=[validators.InputRequired()]
    )

    end = DateField(
        label=_("End"),
        validators=[validators.InputRequired()]
    )

    def validate(self):
        result = super().validate()

        if self.start.data and self.end.data:
            if self.start.data > self.end.data:
                message = _("The end date must be later than the start date")
                self.end.errors.append(message)
                result = False

        return result


class ExportForm(Form):
    """ A form providing a choice of export formats. """

    file_format = RadioField(
        label=_("Format"),
        choices=[
            ('csv', _("CSV File")),
            ('xlsx', _("Excel File")),
            ('json', _("JSON File"))
        ],
        default='csv',
        validators=[
            validators.InputRequired()
        ]
    )

    @property
    def format(self):
        return self.file_format.data

    def as_export_response(self, results, title='export', **kwargs):
        """ Turns the given results (list of dicts) into a webob response
        with the currently selected fiel format.

        The additional keyword arguments are directly passed into the
        convert_list_of_dicts_to_* functions.

        For json, these additional arguments are ignored.

        """

        if self.format == 'json':
            return Response(
                json_body=results,
                content_type='application/json'
            )

        if self.format == 'csv':
            return Response(
                convert_list_of_dicts_to_csv(results, **kwargs),
                content_type='text/plain'
            )

        if self.format == 'xlsx':
            return Response(
                convert_list_of_dicts_to_xlsx(results, **kwargs),
                content_type=(
                    'application/vnd.openxmlformats'
                    '-officedocument.spreadsheetml.sheet'
                ),
                content_disposition='inline; filename={}.xlsx'.format(
                    normalize_for_url(title)
                )
            )

        raise NotImplemented()



from onegov.core.security import Private
from onegov.org import OrgApp, _
from onegov.org.forms import ResourceRecipientForm
from onegov.org.layout import ResourceRecipientsLayout
from onegov.org.layout import ResourceRecipientsFormLayout
from onegov.org.models import ResourceRecipient, ResourceRecipientCollection
from sqlalchemy.orm import undefer
from onegov.org.elements import DeleteLink, Link


@OrgApp.html(
    model=ResourceRecipientCollection,
    template='resource_recipients.pt',
    permission=Private)
def view_resource_recipients(self, request):

    layout = ResourceRecipientsLayout(self, request)

    def recipient_links(recipient):
        yield Link(
            text=_("Edit"),
            url=request.link(recipient, 'bearbeiten')
        )

        yield DeleteLink(
            text=_("Delete"),
            url=layout.csrf_protected_url(request.link(recipient)),
            confirm=_('Do you really want to delete "${name}"?', mapping={
                'name': recipient.name
            }),
            target='#{}'.format(recipient.id.hex),
            yes_button_text=_("Delete Recipient")
        )

    return {
        'layout': layout,
        'title': _("Recipients"),
        'recipients': self.query().options(undefer(ResourceRecipient.content)),
        'recipient_links': recipient_links
    }


@OrgApp.form(
    model=ResourceRecipientCollection,
    name='neuer-empfaenger',
    template='form.pt',
    permission=Private,
    form=ResourceRecipientForm)
def handle_new_resource_recipient(self, request, form):

    if form.submitted(request):
        self.add(
            name=form.name.data,
            medium='email',
            address=form.address.data,
            send_on=form.send_on.data,
            resources=form.resources.data,
        )

        request.success(_("Added a new recipient"))
        return request.redirect(request.link(self))

    title = _("New Recipient")

    return {
        'title': title,
        'layout': ResourceRecipientsFormLayout(self, request, title),
        'form': form
    }


@OrgApp.form(
    model=ResourceRecipient,
    name='bearbeiten',
    template='form.pt',
    permission=Private,
    form=ResourceRecipientForm)
def handle_edit_resource_recipient(self, request, form):

    if form.submitted(request):
        form.populate_obj(self)
        request.success(_("Your changes were saved"))

        return request.redirect(
            request.class_link(ResourceRecipientCollection)
        )
    elif not request.POST:
        form.process(obj=self)

    title = _("Edit Recipient")

    return {
        'title': title,
        'layout': ResourceRecipientsFormLayout(self, request, title),
        'form': form
    }


@OrgApp.view(
    model=ResourceRecipient,
    permission=Private,
    request_method='DELETE')
def delete_notification(self, request):
    request.assert_valid_csrf_token()
    ResourceRecipientCollection(request.app.session()).delete(self)

    @request.after
    def remove_target(response):
        response.headers.add('X-IC-Remove', 'true')
